from typing import TypedDict, List, Literal, Optional, Dict
from dataclasses import dataclass

from ..handler.docker import DockerTask
from ..handler.shell import ShellTask


class Task(TypedDict):
    name: str
    type: Literal["shell", "docker", "group"]
    shell: Optional[ShellTask]
    docker: Optional[DockerTask]
    run_mode: Literal["wait_completion", "indefinite"]
    completion_pattern: Optional[str]
    pattern_in_stderr: Optional[bool]
    depends: Optional[List[str]]


class PaneConfiguration(TypedDict):
    columns: int
    tasks: List[str]


class TasksConfiguration(TypedDict):
    tasks: Dict[str, Task]
    gui: Optional[Dict[str, PaneConfiguration]]


@dataclass
class TaskNode:
    task: Task
    dependencies: List["TaskNode"]




from logging import LogRecord
from typing import Optional

from PyQt6.QtWidgets import QGroupBox, QPlainTextEdit, QVBoxLayout, QSizePolicy, QHBoxLayout, QWidget, \
    QLabel, QLineEdit
from tinyioc import inject

from .data_signals import DataUpdateSignalEmitter
from ..palette.base import BaseColorPalette

from .. import constants


class TaskPanel(QGroupBox):
    _layout: QVBoxLayout
    _actions_group_widget: QWidget
    _actions_group_layout: QVBoxLayout
    _output_stream_edit_text: QPlainTextEdit

    _output_stream: str

    _task_name: str
    _task_label: QLabel

    _filter_edit_text: QLineEdit

    @inject()
    def __init__(self, parent: Optional[QWidget], task_name: str, palette: BaseColorPalette):
        super(TaskPanel, self).__init__(parent)

        self._task_name = task_name
        self._output_stream = ""

        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)

        self.setStyleSheet(f"""
            background-color: {palette.current_line};
            border: 0;
        """)

        self._actions_group_widget = QWidget(self)
        self._actions_group_layout = QVBoxLayout(self)
        self._actions_group_layout.setContentsMargins(0, 0, 0, 0)
        self._actions_group_widget.setLayout(self._actions_group_layout)

        self._layout.addWidget(self._actions_group_widget)

        self._task_label = QLabel(self)
        self._task_label.setText(self._task_name)
        self._task_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self._task_label.setMaximumWidth(80)
        self._task_label.setStyleSheet(f"""
            color: {palette.foreground};
            font-weight: bold;
        """)
        self._actions_group_layout.addWidget(self._task_label)

        self._filter_edit_text = QLineEdit(self)
        self._filter_edit_text.setPlaceholderText("Filter")
        self._filter_edit_text.setStyleSheet(f"""
            background-color: {palette.background};
            color: {palette.foreground};
        """)
        self._filter_edit_text.textChanged.connect(self._filter_changed)
        self._actions_group_layout.addWidget(self._filter_edit_text)

        self._output_stream_edit_text = QPlainTextEdit(self)
        self._output_stream_edit_text.setReadOnly(True)
        self._output_stream_edit_text.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._output_stream_edit_text.setStyleSheet(f"""
            background-color: {palette.background}; 
            color: {palette.foreground}; 
            font-family: monospace;
        """)
        self._layout.addWidget(self._output_stream_edit_text, 1)

    # noinspection PyUnresolvedReferences
    def append_text(self, record: LogRecord):
        scroll_bottom = False

        if self._output_stream_edit_text.verticalScrollBar().value() > \
                self._output_stream_edit_text.verticalScrollBar().maximum() - constants.SCROLL_TOLERANCE:
            scroll_bottom = True

        self._output_stream += record.message
        self._update_output_edit_text()

        if scroll_bottom:
            self._output_stream_edit_text.verticalScrollBar().setValue(
                self._output_stream_edit_text.verticalScrollBar().maximum())

    def _update_output_edit_text(self):
        filter_input = self._filter_edit_text.text()

        if filter_input:
            filtered_text = ""
            for line in self._output_stream.splitlines(keepends=True):
                if filter_input in line:
                    filtered_text += line
        else:
            filtered_text = self._output_stream

        self._output_stream_edit_text.setPlainText(filtered_text)

    def _filter_changed(self):
        self._update_output_edit_text()



import sys
import time
from queue import Queue, Empty
from threading import Thread
from typing import List, Callable, Optional, Dict

from PyQt6.QtWidgets import QApplication

from .main_window import MainWindow
from .. import constants
from ..logger import logger
from ..types.task import PaneConfiguration


class UiApplication:
    _ui_thread: Thread
    _running: bool

    _app: QApplication
    _window: Optional[MainWindow]
    _config: Optional[Dict[str, PaneConfiguration]]

    _task_list: List[str]
    _streams_queue: Queue
    _close_handler: Callable

    _stream_dequeue_thread: Thread
    _stream_dequeue_running: bool

    _trigger_close_handler: bool

    def __init__(self, tasks: List[str], close_handler: Callable, task_streams_queue: Queue,
                 config: Optional[Dict[str, PaneConfiguration]]):
        self._window = None
        self._trigger_close_handler = True
        self._config = config
        self._task_list = tasks
        self._close_handler = close_handler
        self._streams_queue = task_streams_queue

    def _app_quitting(self):
        self._stream_dequeue_running = False
        self._stream_dequeue_thread.join()

        self._close_handler()

    def start_ui(self):
        self._stream_dequeue_running = True
        self._stream_dequeue_thread = Thread(target=self._dequeue_stream)
        self._stream_dequeue_thread.start()

        self._running = True
        self._ui_thread = Thread(target=self._run_ui_thread)
        self._ui_thread.start()

    def _dequeue_stream(self):
        while self._stream_dequeue_running:
            if self._window:
                try:
                    stream_record = self._streams_queue.get(block=True, timeout=0.5)
                    self._window.dispatch_stream_record(stream_record)
                except Empty:
                    pass
            else:
                time.sleep(.1)

    def _run_ui_thread(self):
        self._app = QApplication(sys.argv)
        self._app.setQuitOnLastWindowClosed(True)

        self._window = MainWindow(self._task_list, gui_config=self._config)
        self._window.show()

        self._app.exec()
        self._running = False
        self._app_quitting()

    def stop_ui(self):
        if self._running:
            self._trigger_close_handler = False
            self._window.dispatch_app_termination()

            try:
                self._ui_thread.join(timeout=3)
            except RuntimeError:
                logger.warning("Couldn't stop the UI thread, ignoring it")



#!/usr/bin/env python
import argparse
import asyncio
import logging
import sys
import traceback
from logging.handlers import QueueHandler
from queue import Queue
from typing import List, Set, Dict
from tinyioc import register_instance

from .palette.base import BaseColorPalette
from .palette.darcula import DarculaColorPalette
from .ui.application import UiApplication
from .handler.group import GroupTaskHandler
from .handler.docker import DockerTaskHandler
from .handler.shell import ShellTaskHandler

from .configuration import load_config
from .runner import TaskRunner
from .types.task import Task, TasksConfiguration
from .logger import logger, NewlineStreamHandler

parser = argparse.ArgumentParser(prog="jorun", description="A smart task runner", add_help=True)

parser.add_argument("configuration_file", help="The yml configuration file to run")
parser.add_argument("--level", help="The log level (DEBUG, INFO, ...)", default="INFO", type=str)
parser.add_argument("--file-output", help="Log tasks output to files, one per task. "
                                          "This option lets you specify the directory of the log files", type=str)
parser.add_argument("--gui", help="Force running with the graphical interface", action='store_true')
parser.add_argument("--no-gui", help="Force running without the graphical interface", action='store_true')

running_tasks: List[TaskRunner] = []
async_tasks: Set[asyncio.Task] = set()

missing_tasks: Dict[str, Task] = {}
completed_tasks: Set[str] = set()

program_arguments: argparse.Namespace

task_handlers = [
    ShellTaskHandler(),
    DockerTaskHandler(),
    GroupTaskHandler()
]

task_streams_queue = Queue()

log_handler: logging.Handler

ui_application: UiApplication


def run_missing_tasks():
    tasks_to_run: List[Task] = [t for t in missing_tasks.values() if
                                (not t.get("depends") or len(t["depends"]) == 0) or set(t["depends"]).issubset(
                                    completed_tasks)]

    for task in tasks_to_run:
        missing_tasks.pop(task["name"])

    for task in tasks_to_run:
        run_task(task)


def run_task(task: Task):
    t = TaskRunner(task, task_handlers, program_arguments.file_output, program_arguments.level,
                   log_handler)
    running_tasks.append(t)

    def cb():
        logger.debug(f"Task {task['name']} completed")
        completed_tasks.add(task["name"])

        logger.debug(f"Launching task {task['name']} dependencies")
        run_missing_tasks()

    logger.debug(f"Running task {task['name']}")
    async_t = asyncio.get_event_loop().create_task(t.start(cb))
    async_tasks.add(async_t)
    async_t.add_done_callback(async_tasks.discard)


def main():
    global missing_tasks, program_arguments, ui_application, log_handler

    register_instance(DarculaColorPalette(), register_for=BaseColorPalette)

    program_arguments = parser.parse_args()
    logger.setLevel(program_arguments.level)

    logger.debug("Loading configuration file")
    config: TasksConfiguration = load_config(program_arguments.configuration_file)

    gui_config = config.get("gui")
    show_gui = not program_arguments.no_gui and (program_arguments.gui or gui_config)

    if show_gui:
        logger.debug("Using graphical interface")
        log_handler = QueueHandler(task_streams_queue)
    else:
        logger.debug("Using console output")
        log_handler = NewlineStreamHandler(sys.stdout)

    missing_tasks = config["tasks"].copy()
    loop = asyncio.get_event_loop()

    if show_gui:
        def on_app_stop():
            logger.info("Main window closed")
            loop.stop()

        ui_tasks = [t_name for t_name, t_val in missing_tasks.items() if t_val["type"] != "group"]

        ui_application = UiApplication(ui_tasks, on_app_stop, task_streams_queue, gui_config)
        ui_application.start_ui()

    try:
        run_missing_tasks()
        loop.run_forever()
    except KeyboardInterrupt:
        logger.info("Requested termination")
    except Exception as e:
        logger.error("An error occurred")
        traceback.print_exception(e)
    finally:
        logger.info("Terminating the async loop...")
        if loop.is_running():
            loop.stop()

        if show_gui:
            logger.info("Killing gui...")
            ui_application.stop_ui()

        logger.info("Killing async tasks...")
        for t in async_tasks:
            t.cancel()

        logger.info("Killing running tasks...")
        for i in reversed(range(len(running_tasks))):
            t = running_tasks[i]
            t.stop()
            running_tasks.pop(i)


if __name__ == "__main__":
    main()



import os
import subprocess
import asyncio
from asyncio.subprocess import Process
from typing import Callable, Optional, List, Union, Dict

from ..types.options import TaskOptions
from .base import BaseTaskHandler
from ..logger import logger


class ShellTask(TaskOptions):
    command: Union[str, List[str]]
    working_directory: Optional[str]
    environment: Optional[Dict[str, str]]


class ShellTaskHandler(BaseTaskHandler):
    @property
    def task_type(self) -> str:
        return "shell"

    async def execute(self, options: Optional[ShellTask], completion_callback: Callable, stderr_redirect: bool) \
            -> Optional[Process]:
        stderr_file = subprocess.STDOUT if stderr_redirect else subprocess.PIPE

        out_cmd = options['command']
        if isinstance(out_cmd, list):
            out_cmd = " ".join(out_cmd)

        logger.debug(f"Running command: {out_cmd}")
        envs = options.get("environment")
        env_vars = None

        if envs:
            env_vars = dict(os.environ)
            env_vars.update({k: str(v) for k, v in envs.items()})

        if not isinstance(options["command"], list):
            process = await asyncio.create_subprocess_shell(
                options["command"],
                cwd=options.get("working_directory"),
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=stderr_file,
                stdin=subprocess.DEVNULL)
        else:
            process = await asyncio.create_subprocess_exec(
                options["command"][0],
                *options["command"][1:],
                cwd=options.get("working_directory"),
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=stderr_file,
                stdin=subprocess.DEVNULL)

        return process

    def on_exit(self, options: TaskOptions, process: Process):
        pass



from logging import LogRecord
from typing import Optional, List, Dict

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QWidget, QSplitter, QVBoxLayout

from tinyioc import inject

from ..palette.base import BaseColorPalette
from .task_panel import TaskPanel


class TasksPane(QWidget):
    _layout: QVBoxLayout
    _central_widget: QSplitter
    _tasks: List[str]
    _total_columns: int
    _task_widgets: Dict[str, TaskPanel]
    _splitters: List[QSplitter]

    @inject()
    def __init__(self, parent: Optional[QWidget], tasks: List[str], palette: BaseColorPalette, columns: int = 3):
        super(TasksPane, self).__init__(parent)

        self._tasks = tasks
        self._total_columns = columns
        self._task_widgets = {}
        self._splitters = []

        self._central_widget = QSplitter(Qt.Orientation.Vertical, self)
        self._central_widget.setStyleSheet(f"""
            background-color: {palette.background};
        """)

        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)
        self.setContentsMargins(4, 4, 4, 4)

        self._layout.addWidget(self._central_widget)

        col = 0
        for task in self._tasks:
            if col % self._total_columns == 0:
                last_splitter = QSplitter(Qt.Orientation.Horizontal, self._central_widget)
                last_splitter.setStyleSheet(f"""
                    background-color: {palette.background};
                """)
                self._central_widget.addWidget(last_splitter)
                self._splitters.append(last_splitter)

            task_panel = TaskPanel(self._splitters[-1], task)
            self._splitters[-1].addWidget(task_panel)
            self._task_widgets[task] = task_panel

            col += 1

    # noinspection PyUnresolvedReferences
    def dispatch_log_record(self, record: LogRecord):
        if record.subprocess in self._task_widgets:
            self._task_widgets[record.subprocess].append_text(record)



from dataclasses import dataclass

from .base import BaseColorPalette


@dataclass
class DarculaColorPalette(BaseColorPalette):
    background: str = "#282a36"
    current_line: str = "#44475a"
    selection: str = "#44475a"
    foreground: str = "#f8f8f2"
    comment: str = "#6272a4"



from PyQt6.QtCore import QObject, pyqtSignal


class MainWindowSignals(QObject):
    data_received = pyqtSignal(object)
    app_terminated = pyqtSignal()


class DataUpdateSignalEmitter:
    signals: MainWindowSignals



from logging import LogRecord
from typing import Dict, List, Optional

from PyQt6.QtWidgets import QMainWindow, QTabWidget

from tinyioc import inject

from .data_signals import DataUpdateSignalEmitter, MainWindowSignals
from .pane import TasksPane
from .. import constants
from ..palette.base import BaseColorPalette
from ..types.task import PaneConfiguration


class MainWindow(QMainWindow, DataUpdateSignalEmitter):
    _tab_widget: QTabWidget
    _panes: List[TasksPane]
    signals = MainWindowSignals()

    @inject()
    def __init__(self, tasks: List[str], palette: BaseColorPalette, gui_config: Optional[Dict[str, PaneConfiguration]]):
        super(MainWindow, self).__init__()

        self.signals.app_terminated.connect(self.close)
        self._panes = []

        self.setStyleSheet(f"""
            background-color: {palette.background};
        """)
        self.setMinimumSize(300, 300)

        self.setWindowTitle(constants.APP_NAME)

        self._tab_widget = QTabWidget(self)
        self._tab_widget.setStyleSheet(f"""
            QTabWidget::pane
            {{
                background: {palette.selection};
            }}
            
            QTabBar::tab {{
                background: {palette.selection};
                color: {palette.foreground};
                border-top-left-radius: 8px;
                border-top-right-radius: 8px;
                padding: 6px;
                margin: 0 1px 0 1px;
            }}
            
            QTabBar::tab:selected {{
                font-weight: bold;
            }}
        """)

        self.setCentralWidget(self._tab_widget)

        tasks_bucket = set(tasks)

        if gui_config:
            for pane_name, pane_options in gui_config.items():
                pane = TasksPane(None, pane_options.get("tasks"), columns=pane_options.get("columns") or 3)
                pane.setAutoFillBackground(True)
                self._tab_widget.addTab(pane, pane_name)
                self._panes.append(pane)

                for t in pane_options["tasks"]:
                    tasks_bucket.discard(t)

        if len(tasks_bucket) > 0:
            extra_pane = TasksPane(None, list(tasks_bucket))
            extra_pane.setAutoFillBackground(True)
            self._tab_widget.addTab(extra_pane, "-")
            self._panes.append(extra_pane)

        self.signals.data_received.connect(self._handle_stream_record)

    def _handle_stream_record(self, record: LogRecord):
        for p in self._panes:
            p.dispatch_log_record(record)

    # noinspection PyUnresolvedReferences
    def dispatch_stream_record(self, record: LogRecord):
        self.signals.data_received.emit(record)

    def dispatch_app_termination(self):
        self.signals.app_terminated.emit()
